# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import math
import random
import re
import string
import textwrap
from re import Match

import emoji
from gi.repository import GLib
from nbxmpp.structs import LocationData
from nbxmpp.structs import TuneData

from gajim.common import regex
from gajim.common.const import LOCATION_DATA
from gajim.common.i18n import _
from gajim.common.regex import NON_SPACING_MARKS_REGEX
from gajim.common.regex import URL_REGEX

# from RFC 3986, 3.3. Path (pchar without unreserved and pct-encoded):
_reserved_chars_allowed_in_path_segment = regex.sub_delims + ":@"


def remove_invalid_xml_chars(input_string: str) -> str:
    if input_string:
        return re.sub(regex.INVALID_XML_CHARS_REGEX, "", input_string)
    return input_string


def escape_iri_path_segment(s: str) -> str:
    return GLib.Uri.escape_string(s, _reserved_chars_allowed_in_path_segment, True)


def escape_iri_path(s: str) -> str:
    return GLib.Uri.escape_string(
        s, _reserved_chars_allowed_in_path_segment + "/", True
    )


def escape_iri_query(s: str) -> str:
    return GLib.Uri.escape_string(
        s, _reserved_chars_allowed_in_path_segment + "/?", True
    )


def jid_to_iri(jid: str) -> str:
    return "xmpp:" + escape_iri_path(jid)


def quote_text(text: str) -> str:
    return "> " + text.replace("\n", "\n> ") + "\n"


def format_duration(ns: float, total_ns: float) -> str:
    seconds = ns / 1e9
    minutes = seconds / 60
    hours = minutes / 60

    total_minutes = total_ns / 1e9 / 60
    total_hours = total_minutes / 60

    i_seconds = int(seconds) % 60
    i_minutes = int(minutes) % 60
    i_hours = int(hours)

    if total_hours >= 1:
        width = len(str(int(total_hours)))
        return (f"%0{width}d" % i_hours) + f":{i_minutes:02d}:{i_seconds:02d}"

    if total_minutes >= 1:
        width = len(str(int(total_minutes)))
        return (f"%0{width}d" % i_minutes) + f":{i_seconds:02d}"

    return f"0:{i_seconds:02d}"


def format_eta(time_: int | float) -> str:
    times = {"minutes": 0, "seconds": 0}
    time_ = int(time_)
    times["seconds"] = time_ % 60
    if time_ >= 60:
        time_ = int(time_ / 60)
        times["minutes"] = round(time_ % 60)
        return _("%(minutes)s min %(seconds)s s") % times
    return _("%s s") % times["seconds"]


def format_fingerprint(fingerprint: str) -> str:
    fplen = len(fingerprint)
    wordsize = fplen // 8
    buf = ""
    for char in range(0, fplen, wordsize):
        buf += f"{fingerprint[char : char + wordsize]} "
    buf = textwrap.fill(buf, width=36)
    return buf.rstrip().upper()


def format_tune(data: TuneData) -> str:
    artist = GLib.markup_escape_text(data.artist or _("Unknown Artist"))
    title = GLib.markup_escape_text(data.title or _("Unknown Title"))
    source = GLib.markup_escape_text(data.source or _("Unknown Source"))

    return _('<b>"%(title)s"</b> by <i>%(artist)s</i>\nfrom <i>%(source)s</i>') % {
        "title": title,
        "artist": artist,
        "source": source,
    }


def format_location(location: LocationData) -> str:
    location_dict = location._asdict()
    location_string = ""
    for attr, value in location_dict.items():
        if value is None:
            continue
        text = GLib.markup_escape_text(value)
        # Translate standard location tag
        tag = LOCATION_DATA.get(attr)
        if tag is None:
            continue
        location_string += f"\n<b>{tag.capitalize()}</b>: {text}"

    return location_string.strip()


def make_href_markup(string: str | None) -> str:
    if not string:
        return ""

    string = process_non_spacing_marks(string)
    string = GLib.markup_escape_text(string)

    def _to_href(match: Match[str]) -> str:
        url = match.group()
        if "://" not in url:
            url = f"https://{url}"
        return f'<a href="{url}">{match.group()}</a>'

    return URL_REGEX.sub(_to_href, string)


def format_bytes_as_hex(bytes_: bytes, line_count: int = 1) -> str:
    line_length = math.ceil(len(bytes_) / line_count)

    hex_list = [f"{b:02X}" for b in bytes_]

    lines: list[str] = []
    for pos in range(0, len(bytes_), line_length):
        lines.append(":".join(hex_list[pos : pos + line_length]))
    return "\n".join(lines)


def process_non_spacing_marks(string: str) -> str:
    """
    Helper function for working around unicode non-spacing marks in
    conjunction with Pango.WrapMode.WORD_CHAR, see:
    https://gitlab.gnome.org/GNOME/pango/-/issues/798
    Unbreaks spaces around non-spacing marks.
    """
    return NON_SPACING_MARKS_REGEX.sub("\u00a0", string)


def normalize_reactions(reactions: list[str]) -> tuple[set[str], set[str]]:
    valid: set[str] = set()
    invalid: set[str] = set()
    # Set arbitrary limit of max reactions to prevent
    # performance problems when loading and displaying them.
    reactions = reactions[:10]
    for reaction in reactions:
        # Remove emoji variant selectors. They are not needed because
        # reactions are required to be shown as emoji representation.
        # Furthermore it allows us to unify both versions.
        reaction = reaction.strip("\ufe0e\ufe0f")
        if not emoji.is_emoji(reaction):
            invalid.add(reaction)
            continue
        valid.add(reaction)

    return valid, invalid


def convert_to_codepoints(string: str) -> str:
    return "".join(f"\\u{ord(c):04x}" for c in string)


def get_country_flag_from_code(country_code: str) -> str:
    """Returns a flag emoji for a two-letter country code."""
    emoji_letters = {
        "A": "🇦",
        "B": "🇧",
        "C": "🇨",
        "D": "🇩",
        "E": "🇪",
        "F": "🇫",
        "G": "🇬",
        "H": "🇭",
        "I": "🇮",
        "J": "🇯",
        "K": "🇰",
        "L": "🇱",
        "M": "🇲",
        "N": "🇳",
        "O": "🇴",
        "P": "🇵",
        "Q": "🇶",
        "R": "🇷",
        "S": "🇸",
        "T": "🇹",
        "U": "🇺",
        "V": "🇻",
        "W": "🇼",
        "X": "🇽",
        "Y": "🇾",
        "Z": "🇿",
    }

    if len(country_code) != 2:
        return country_code.upper()

    first = emoji_letters.get(country_code[0].upper())
    second = emoji_letters.get(country_code[1].upper())
    if first is None or second is None:
        return country_code.upper()

    return f"{first}{second}"


def to_one_line(msg: str) -> str:
    return " ".join(msg.splitlines())


def from_one_line(msg: str) -> str:
    # (?<!\\) is a lookbehind assertion which asks anything but '\'
    # to match the regexp that follows it

    # So here match '\\n' but not if you have a '\' before that
    expr = re.compile(r"(?<!\\)\\n")
    msg = expr.sub("\n", msg)
    return msg.replace("\\\\", "\\")


def get_random_string(count: int = 16) -> str:
    """
    Create random string of length 'count'

    WARNING: Don't use this for security purposes
    """
    allowed = string.ascii_uppercase + string.digits
    return "".join(random.choice(allowed) for _char in range(count))



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# XEP-0258: Security Labels in XMPP

from __future__ import annotations

from typing import Any

from collections.abc import Generator

from nbxmpp.errors import is_error
from nbxmpp.modules.security_labels import Catalog
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.structs import DiscoInfo

from gajim.common import app
from gajim.common import types
from gajim.common.events import SecCatalogReceived
from gajim.common.modules.base import BaseModule
from gajim.common.modules.util import as_task


class SecLabels(BaseModule):

    _nbxmpp_extends = 'SecurityLabels'
    _nbxmpp_methods = [
        'request_catalog',
    ]

    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)

        self._catalogs: dict[JID, Catalog] = {}
        self.supported = False

    def pass_disco(self, info: DiscoInfo) -> None:
        if Namespace.SECLABEL not in info.features:
            return

        self.supported = True
        self._log.info('Discovered security labels: %s', info.jid)

    @as_task
    def request_catalog(self, jid: JID) -> Generator[Any, Any]:
        _task = yield

        catalog = yield self._nbxmpp('SecurityLabels').request_catalog(str(jid))

        if is_error(catalog):
            self._log.info(catalog)
            return

        assert isinstance(catalog, Catalog)
        self._catalogs[jid] = catalog

        self._log.info('Received catalog: %s', jid)

        app.ged.raise_event(SecCatalogReceived(account=self._account,
                                               jid=jid,
                                               catalog=catalog))

    def get_catalog(self, jid: JID) -> Catalog | None:
        if not self.supported:
            return None

        catalog = self._catalogs.get(jid)
        if catalog is None:
            self.request_catalog(jid)
            return None
        return catalog



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only


from __future__ import annotations

from typing import Any
from typing import cast
from typing import Literal
from typing import TYPE_CHECKING

import datetime
import gc
import logging
import os
import pprint
import sys
import weakref
import xml.etree.ElementTree as ET
from collections.abc import Iterator
from functools import partial
from functools import wraps
from pathlib import Path

import cairo
from gi.repository import Adw
from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Pango
from packaging.requirements import Requirement

from gajim.common import app
from gajim.common import types
from gajim.common.configpaths import get_ui_path
from gajim.common.const import AvatarSize
from gajim.common.const import LINUX_EXECUTABLE_EXTENSIONS
from gajim.common.const import MACOS_EXECUTABLE_EXTENSIONS
from gajim.common.const import WINDOWS_EXECUTABLE_EXTENSIONS
from gajim.common.i18n import _
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.modules.contacts import GroupchatParticipant
from gajim.common.storage.archive import models as mod
from gajim.common.storage.archive.const import ChatDirection
from gajim.common.storage.archive.const import MessageType
from gajim.common.styling import PlainBlock
from gajim.common.util.decorators import catch_exceptions
from gajim.common.util.uri import geo_provider_from_location
from gajim.common.util.uri import GeoUri
from gajim.common.util.uri import InvalidUri
from gajim.common.util.uri import parse_uri
from gajim.common.util.uri import Uri
from gajim.common.util.uri import UriT
from gajim.common.util.uri import XmppIri

from gajim.gtk.alert import ConfirmationAlertDialog
from gajim.gtk.const import GDK_MEMORY_DEFAULT

if sys.platform == "darwin" or TYPE_CHECKING:
    from AppKit import NSWorkspace
    from Foundation import NSURL

log = logging.getLogger("gajim.gtk.util")


def get_gtk_version() -> str:
    return "%i.%i.%i" % (
        Gtk.get_major_version(),
        Gtk.get_minor_version(),
        Gtk.get_micro_version(),
    )


def get_adw_version() -> str:
    return "%i.%i.%i" % (
        Adw.get_major_version(),
        Adw.get_minor_version(),
        Adw.get_micro_version(),
    )


def gi_gui_package_version(requirement: str):
    req = Requirement(requirement)
    match req.name:
        case "Gtk":
            version = get_gtk_version()
        case "Adw":
            version = get_adw_version()
        case _:
            raise ValueError("Unknown lib name: %s" % req.name)

    return req.specifier.contains(version)


def scroll_to(widget: Gtk.ScrolledWindow, pos: Literal["top", "bottom"]) -> bool:
    """Scrolls to `pos` of a GtkScrolledWindow.

    Args:
        widget (GtkScrolledWindow)

    Returns:
        bool: The return value is False so it can be used with GLib.idle_add.
    """
    adj_v = widget.get_vadjustment()
    if adj_v is None:  # pyright: ignore
        # This can happen when the Widget is already destroyed when called
        # from GLib.idle_add
        return False

    scroll_pos = 0
    if pos == "bottom":
        scroll_pos = adj_v.get_upper() - adj_v.get_page_size()
    adj_v.set_value(scroll_pos)

    adj_h = widget.get_hadjustment()
    adj_h.set_value(0)
    return False


def at_the_end(widget: Gtk.ScrolledWindow) -> bool:
    """Determines if a Scrollbar in a GtkScrolledWindow is at the end.

    Args:
        widget (GtkScrolledWindow)

    Returns:
        bool: The return value is True if at the end, False if not.
    """
    adj_v = widget.get_vadjustment()
    max_scroll_pos = adj_v.get_upper() - adj_v.get_page_size()
    return adj_v.get_value() == max_scroll_pos


def ensure_not_destroyed(func: Any) -> Any:
    @wraps(func)
    def func_wrapper(self: Any, *args: Any, **kwargs: Any):
        if self._destroyed:  # pylint: disable=protected-access
            return None
        return func(self, *args, **kwargs)

    return func_wrapper


def get_contact_name_for_message(
    db_row: mod.Message, contact: types.ChatContactT
) -> str:
    if isinstance(contact, BareContact) and contact.is_self:
        return _("Me")

    if db_row.type == MessageType.CHAT:
        if db_row.direction == ChatDirection.INCOMING:
            return contact.name
        return app.nicks[contact.account]

    elif db_row.type == MessageType.GROUPCHAT:
        resource = db_row.resource
        if resource is None:
            # Fall back to MUC name if contact name is None
            # (may be the case for service messages from the MUC)
            return contact.name
        return resource

    elif db_row.type == MessageType.PM:
        if occupant := db_row.occupant:
            if name := occupant.nickname:
                return name

        if db_row.resource is None:
            if db_row.direction == ChatDirection.OUTGOING:
                return _("Me")
            # This should never happen better degrade gracefully
            return "Unknown"
        return db_row.resource

    else:
        raise ValueError


def get_avatar_for_message(
    db_row: mod.Message, contact: types.ChatContactT, scale: int, size: AvatarSize
) -> Gdk.Texture | None:
    if db_row.direction == ChatDirection.OUTGOING:
        return app.app.avatar_storage.get_own_avatar_texture(
            contact.account, size, scale
        )

    match db_row.type:
        case MessageType.GROUPCHAT | MessageType.PM:
            if db_row.resource is None:
                # Message from Groupchat itself
                assert isinstance(contact, GroupchatContact)
                return contact.get_avatar(size, scale)

            return app.app.avatar_storage.get_occupant_texture(
                db_row.remote.jid, db_row.occupant or db_row.resource, size, scale
            )

        case MessageType.CHAT:
            if isinstance(contact, BareContact):
                return contact.get_avatar(size, scale, add_show=False)
            log.warning("Unexpected contact type %r, unable to get avatar", contact)
            return None

        case _:
            raise ValueError(f"Unhandled type: {db_row.type}")


def make_pango_attributes(block: PlainBlock) -> Pango.AttrList:
    attrlist = Pango.AttrList()
    for span in block.spans:
        attr = get_style_attribute_with_name(span.name)
        attr.start_index = span.start_byte
        attr.end_index = span.end_byte
        attrlist.insert(attr)
    return attrlist


_grapheme_buffer = Gtk.TextBuffer()


def get_first_graphemes(text: str, n: int) -> str:
    # This should be possible with lower-level APIs like Pango.break_* or
    # Pango.get_log_attrs, but their Python bindings seem totally broken.
    # The reuse of one global buffer is to mitigate very probable memory leaks.
    _grapheme_buffer.set_text(text)
    cursor = _grapheme_buffer.get_start_iter()
    cursor.forward_cursor_positions(n)
    return _grapheme_buffer.get_slice(_grapheme_buffer.get_start_iter(), cursor, False)


def get_first_grapheme(text: str) -> str:
    return get_first_graphemes(text, 1)


def get_style_attribute_with_name(name: str) -> Pango.Attribute:
    if name == "strong":
        return Pango.attr_weight_new(Pango.Weight.BOLD)

    if name == "strike":
        return Pango.attr_strikethrough_new(True)

    if name == "emphasis":
        return Pango.attr_style_new(Pango.Style.ITALIC)

    if name == "pre":
        return Pango.attr_family_new("monospace")

    raise ValueError("unknown attribute %s" % name)


def get_key_theme() -> str | None:
    settings = Gtk.Settings.get_default()
    if settings is None:
        return None
    return settings.get_property("gtk-key-theme-name")


def allow_send_message(has_text: bool, contact: types.ChatContactT) -> bool:
    if isinstance(contact, GroupchatContact):
        joined = contact.is_joined

        is_visitor = False
        if joined:
            self_contact = contact.get_self()
            assert self_contact
            is_visitor = self_contact.role.is_visitor

        return bool(has_text and joined and not is_visitor)

    if isinstance(contact, GroupchatParticipant):
        if not contact.is_available:
            return False

        groupchat_contact = contact.room
        joined = groupchat_contact.is_joined

        is_visitor = False
        if joined:
            self_contact = groupchat_contact.get_self()
            assert self_contact
            is_visitor = self_contact.role.is_visitor

        return bool(has_text and joined and not is_visitor)

    # BareContact
    online = app.account_is_connected(contact.account)
    return bool(online and has_text)


def iterate_listbox_children(listbox: Gtk.ListBox) -> Iterator[Gtk.Widget]:
    index = 0
    while child := listbox.get_row_at_index(index):
        yield child
        index += 1


def container_remove_all(container: Any) -> None:
    while child := container.get_first_child():
        container.remove(child)


def clear_listbox(listbox: Gtk.ListBox) -> None:
    """Gtk.ListBox.remove_all() does not work if backed by a model."""
    for row in iterate_listbox_children(listbox):
        listbox.remove(row)


def get_listbox_row_count(listbox: Gtk.ListBox) -> int:
    return len(list(iterate_listbox_children(listbox)))


def iterate_children(widget: Gtk.Widget) -> Iterator[Gtk.Widget]:
    child = widget.get_first_child()
    if child is None:
        return
    yield child

    while child := child.get_next_sibling():
        yield child


def iterate_widget_tree(
    widget: Gtk.Widget, parent: str | None = None, only_children: bool = False
) -> Iterator[tuple[Gtk.Widget, str | None]]:
    if not only_children:
        yield widget, parent

    if parent is None:
        parent = widget.__class__.__name__
    else:
        parent += f" -> {widget.__class__.__name__}"

    child = widget.get_first_child()
    if child is None:
        return

    yield from iterate_widget_tree(child, parent)

    while child := child.get_next_sibling():
        yield from iterate_widget_tree(child, parent)


def convert_surface_to_texture(surface: cairo.ImageSurface) -> Gdk.Texture:
    memoryv = surface.get_data()
    assert memoryv is not None
    return Gdk.MemoryTexture.new(
        width=surface.get_width(),
        height=surface.get_height(),
        format=GDK_MEMORY_DEFAULT,
        bytes=GLib.Bytes.new(memoryv.tobytes()),
        stride=surface.get_stride(),
    )


def convert_texture_to_surface(texture: Gdk.Texture) -> cairo.ImageSurface:
    downloader = Gdk.TextureDownloader.new(texture)
    gbytes, stride = downloader.download_bytes()
    bytes_ = gbytes.get_data()
    assert bytes_ is not None
    array = bytearray(bytes_)

    surface = cairo.ImageSurface.create_for_data(
        array,
        cairo.Format.ARGB32,
        texture.get_width(),
        texture.get_height(),
        stride,
    )
    return surface


def convert_py_to_glib_datetime(dt: datetime.datetime | datetime.date) -> GLib.DateTime:
    if type(dt) is datetime.date:
        dt = datetime.datetime(
            year=dt.year,
            month=dt.month,
            day=dt.day,
            hour=0,
            minute=0,
            second=0,
            tzinfo=datetime.UTC,
        )

    g_dt = GLib.DateTime.new_from_iso8601(dt.isoformat())
    assert g_dt is not None
    return g_dt


def get_ui_string(filename: str) -> bytes:
    path = get_ui_path(filename)

    if sys.platform != "win32":
        return path.read_bytes()

    tree = ET.parse(path)
    for node in tree.findall(".//*[@translatable='yes']"):
        node.text = _(node.text) if node.text else ""
        del node.attrib["translatable"]
    return ET.tostring(tree.getroot(), method="xml")


@catch_exceptions
def open_uri(uri: UriT | str) -> None:
    if isinstance(uri, str):
        uri = parse_uri(uri)

    match uri:
        case InvalidUri():
            log.warning("Failed to open uri: %s", uri.error)

        case XmppIri():
            app.window.open_xmpp_iri(uri)

        case GeoUri():
            if Gio.AppInfo.get_default_for_uri_scheme("geo"):
                open_uri_externally(uri.uri)
            else:
                open_uri(geo_provider_from_location(uri.lat, uri.lon))

        case Uri():
            open_uri_externally(uri.uri)


def open_uri_externally(uri: str) -> None:
    launcher = Gtk.UriLauncher(uri=uri)
    try:
        launcher.launch(app.window)
    except Exception as error:
        log.warning(
            "open_uri_externally: Couldn't launch default for %s: %s", uri, error
        )


def is_file_extension_executable(path: Path) -> bool:
    if not path.suffix:
        return False

    match sys.platform:
        case "win32":
            return path.suffix[1:] in WINDOWS_EXECUTABLE_EXTENSIONS
        case "darwin":
            return path.suffix[1:] in MACOS_EXECUTABLE_EXTENSIONS
        case "linux":
            return path.suffix[1:] in LINUX_EXECUTABLE_EXTENSIONS
        case _:
            return False


@catch_exceptions
def open_file(path: Path, *, show_in_folder: bool = False) -> None:
    if not path.exists():
        log.warning("Unable to open file, path %s does not exist", path)
        return

    path_absolute = path.resolve()

    def _on_open_confirm() -> None:
        launcher.launch(app.window)

    launcher = Gtk.FileLauncher(file=Gio.File.new_for_path(str(path_absolute)))
    try:
        if show_in_folder:
            if (
                sys.platform == "darwin"
            ):  # Implemented as workaround for https://gitlab.gnome.org/GNOME/gtk/-/issues/8070
                assert NSWorkspace is not None and NSURL is not None
                url = NSURL.fileURLWithPath_(str(path_absolute))
                NSWorkspace.sharedWorkspace().activateFileViewerSelectingURLs_([url])
            else:
                launcher.open_containing_folder(app.window)
        else:
            if is_file_extension_executable(path_absolute):
                ConfirmationAlertDialog(
                    _("Open File?"),
                    _(
                        "Opening this file may be harmful for your computer. Do you really want to open this file?"
                    ),
                    confirm_label=_("_Open"),
                    appearance="destructive",
                    callback=_on_open_confirm,
                )
            else:
                launcher.launch(app.window)
    except Exception as error:
        log.warning("Unable to open file: %s", error)


def remove_css_class(widget: Gtk.Widget, css_class_name: str) -> None:
    for css_class in list(widget.get_css_classes()):
        if css_class_name in css_class:
            widget.remove_css_class(css_class)


class ObjTracker:
    def __init__(self) -> None:
        self._gobjects: list[str] = []
        self._pyobjects: dict[str, Any] = {}

    def get_count(self) -> int:
        return len(self._gobjects)

    def track(self, obj: GObject.Object, path: str | None) -> None:
        if path is None:
            full_path = f"({id(obj)}) {obj.__class__.__name__}"
        else:
            full_path = f"({id(obj)}) {path} -> {obj.__class__.__name__}"

        obj.weak_ref(partial(self._gobjects.remove, full_path))
        self._gobjects.append(full_path)

        finalizer = weakref.finalize(obj, partial(self._pyobjects.pop, full_path))
        self._pyobjects[full_path] = finalizer

    def iter_pyobjects(self) -> Iterator[tuple[str, Any]]:
        yield from self._pyobjects.items()

    def iter_gobjects(self) -> Iterator[str]:
        yield from self._gobjects

    def pyobj_finalized(self) -> bool:
        return not bool(self._pyobjects)

    def gobj_finalized(self) -> bool:
        return not bool(self._gobjects)

    def get_pyobject_leaks(self) -> str:
        def is_finalizer_ref(ref: Any) -> bool:
            try:
                return isinstance(ref[2][0], str)
            except Exception:
                return False

        log_string = ""
        for path, finalizer in self.iter_pyobjects():
            tup = finalizer.peek()
            if tup is None:
                continue

            log_string += f"{path} not finalized\n"
            log_string += "References:"

            for ref in gc.get_referrers(tup[0]):
                if is_finalizer_ref(ref):
                    continue
                if isinstance(ref, dict):
                    ref = cast(Any, ref)
                    log_string += f"{pprint.pformat(ref)}\n"
                else:
                    log_string += f"{ref}\n"

        return log_string

    def get_gobject_leaks(self) -> str:
        return "\n".join(self.iter_gobjects())


def check_finalize(obj: GObject.Object, only_children: bool = False) -> None:
    if "GAJIM_LEAK" not in os.environ:
        return

    logger = logging.getLogger("gajim.leak")
    parent_obj = obj.__class__.__name__

    if isinstance(obj, Gtk.Widget):
        objs = iterate_widget_tree(obj, only_children=only_children)
    else:
        objs = [(obj, None)]

    tracker = ObjTracker()

    for _obj, path in objs:
        tracker.track(_obj, path)
    del objs

    obj_count = tracker.get_count()
    if not obj_count:
        return

    def check_finalized():
        gc.collect()

        if tracker.pyobj_finalized():
            logger.info(
                "%s (Python Object) successful finalized (total: %s)",
                parent_obj,
                obj_count,
            )
        else:
            logger.warning(
                "%s (Python Object) failed to finalize (total: %s)\n%s",
                parent_obj,
                obj_count,
                tracker.get_pyobject_leaks(),
            )

        if tracker.gobj_finalized():
            logger.info(
                "%s (GObject) successful finalized (total: %s)", parent_obj, obj_count
            )
        else:
            logger.warning(
                "%s (GObject) failed to finalize (total: %s)\n%s",
                parent_obj,
                obj_count,
                tracker.get_gobject_leaks(),
            )

    GLib.timeout_add_seconds(2, check_finalized)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import typing

import logging
import sys
from pathlib import Path

from gi.repository import Gtk

if sys.platform == 'win32' or typing.TYPE_CHECKING:
    import winsound

if sys.platform == 'darwin' or typing.TYPE_CHECKING:
    from AppKit import NSSound

log = logging.getLogger('gajim.c.sound')


class PlaySound:

    def play(self, path: Path, loop: bool = False) -> None:
        raise NotImplementedError

    def stop(self) -> None:
        raise NotImplementedError

    def loop_in_progress(self) -> bool:
        raise NotImplementedError


class PlatformWindows(PlaySound):

    def __init__(self) -> None:
        self._loop_in_progress = False

    def play(self, path: Path, loop: bool = False) -> None:
        if self._loop_in_progress:
            return

        assert winsound is not None
        flags = (winsound.SND_FILENAME |
                 winsound.SND_ASYNC |
                 winsound.SND_NODEFAULT)
        if loop:
            self._loop_in_progress = True
            flags = flags | winsound.SND_LOOP

        try:
            winsound.PlaySound(str(path), flags)
        except Exception:
            log.exception('Sound Playback Error')

    def stop(self) -> None:
        assert winsound is not None
        try:
            winsound.PlaySound(None, 0)
        except Exception:
            log.exception('Sound Playback Error')
        self._loop_in_progress = False

    def loop_in_progress(self) -> bool:
        return self._loop_in_progress


class PlatformMacOS(PlaySound):

    def play(self, path: Path, loop: bool = False) -> None:
        assert NSSound is not None
        sound = NSSound.alloc()
        sound.initWithContentsOfFile_byReference_(str(path), True)
        sound.play()

    def stop(self) -> None:
        pass

    def loop_in_progress(self) -> bool:
        return False


class PlatformUnix(PlaySound):

    def __init__(self) -> None:
        self._media_file: Gtk.MediaFile | None = None

    def play(self, path: Path, loop: bool = False) -> None:
        if self.loop_in_progress():
            return

        self._media_file = Gtk.MediaFile.new_for_filename(str(path))

        if self._media_file.is_seekable():
            self._media_file.set_loop(loop)

        self._media_file.play()

    def stop(self) -> None:
        if self._media_file is None:
            return

        self._media_file.clear()

    def loop_in_progress(self) -> bool:
        return self._media_file is not None and self._media_file.get_loop()


def _init_platform() -> PlaySound:
    if sys.platform == 'win32':
        return PlatformWindows()

    if sys.platform == 'darwin':
        return PlatformMacOS()

    return PlatformUnix()


_platform_player = _init_platform()


def play(path: Path, loop: bool = False) -> None:
    _platform_player.play(path, loop)


def stop():
    _platform_player.stop()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# XEP-0313: Message Archive Management

from __future__ import annotations

from typing import Any

from collections.abc import Generator
from datetime import datetime
from datetime import timedelta
from datetime import UTC

import nbxmpp
from nbxmpp.errors import is_error
from nbxmpp.errors import MalformedStanzaError
from nbxmpp.errors import StanzaError
from nbxmpp.modules.util import raise_if_error
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.protocol import Message
from nbxmpp.structs import DiscoInfo
from nbxmpp.structs import MAMQueryData
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import StanzaHandler
from nbxmpp.structs import StanzaIDData
from nbxmpp.task import Task
from nbxmpp.util import generate_id

from gajim.common import app
from gajim.common import types
from gajim.common.const import ArchiveState
from gajim.common.const import ClientState
from gajim.common.const import SyncThreshold
from gajim.common.events import ArchivingIntervalFinished
from gajim.common.events import FeatureDiscovered
from gajim.common.events import RawMamMessageReceived
from gajim.common.modules.base import BaseModule
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.modules.util import as_task
from gajim.common.storage.archive import models as mod
from gajim.common.util.datetime import FIRST_UTC_DATETIME


class MAM(BaseModule):

    _nbxmpp_extends = 'MAM'
    _nbxmpp_methods = [
        'request_preferences',
        'set_preferences',
        'make_query',
    ]

    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)

        self.handlers = [
            StanzaHandler(name='message',
                          callback=self._set_message_archive_info,
                          priority=42),
            StanzaHandler(name='message',
                          callback=self._mam_message_received,
                          priority=41),
        ]

        self.available = False
        self._mam_query_ids: dict[JID, str] = {}

        # Holds archive jids where catch up was successful
        self._catch_up_finished: list[JID] = []

        self._con.connect_signal('state-changed', self._on_client_state_changed)
        self._con.connect_signal('resume-failed', self._on_client_resume_failed)

    def pass_disco(self, info: DiscoInfo) -> None:
        if Namespace.MAM_2 not in info.features:
            return

        self.available = True
        self._log.info('Discovered MAM: %s', info.jid)

        app.ged.raise_event(
            FeatureDiscovered(account=self._account,
                              feature=Namespace.MAM_2))

    def _on_client_resume_failed(self,
                                 _client: types.Client,
                                 _signal_name: str
                                 ) -> None:
        self._reset_state()

    def _on_client_state_changed(self,
                                 _client: types.Client,
                                 _signal_name: str,
                                 state: ClientState
                                 ) -> None:
        if state.is_disconnected:
            self._reset_state()

    def _reset_state(self) -> None:
        self._mam_query_ids.clear()
        self._catch_up_finished.clear()

    def _remove_query_by_jid(self, jid: JID) -> None:
        self._mam_query_ids.pop(jid, None)

    def _remove_query_by_id(self, queryid: str) -> None:
        for jid, _queryid in list(self._mam_query_ids.items()):
            if _queryid == queryid:
                del self._mam_query_ids[jid]
                return

    def _is_catch_up_finished(self, jid: JID) -> bool:
        return jid in self._catch_up_finished

    def is_catch_up_finished(self, contact: types.ChatContactT) -> bool:
        if isinstance(contact, BareContact):
            archive_jid = self._get_own_bare_jid()

        elif isinstance(contact, GroupchatContact):
            archive_jid = contact.jid

        else:
            # GroupChatParticipant
            archive_jid = contact.room.jid

        return self._is_catch_up_finished(archive_jid)

    def _from_valid_archive(self,
                            _stanza: Message,
                            properties: MessageProperties
                            ) -> bool:
        if properties.type.is_groupchat:
            assert properties.jid is not None
            expected_archive = properties.jid
        else:
            expected_archive = self._con.get_own_jid()

        assert properties.mam is not None
        return properties.mam.archive.bare_match(expected_archive)

    @staticmethod
    def _get_stanza_id(properties: MessageProperties,
                       archive_jid: JID
                       ) -> StanzaIDData | None:
        for stanza_id in properties.stanza_ids:
            if stanza_id.by == str(archive_jid):
                return stanza_id
        return None

    def _set_message_archive_info(self,
                                  _con: types.NBXMPPClient,
                                  _stanza: Message,
                                  properties: MessageProperties
                                  ) -> None:
        if (properties.is_mam_message or
                properties.is_pubsub or
                properties.is_muc_subject):
            return

        if properties.type.is_groupchat:
            assert properties.jid is not None
            archive_jid = properties.jid.new_as_bare()
            timestamp = properties.timestamp

            disco_info = app.storage.cache.get_last_disco_info(archive_jid)
            if disco_info is None:
                # This is the case on MUC creation
                # After MUC configuration we receive a configuration change
                # message before we had the chance to disco the new MUC
                return

            if disco_info.mam_namespace != Namespace.MAM_2:
                return

        else:
            if not self.available:
                return

            archive_jid = self._con.get_own_jid().new_as_bare()
            timestamp = None

        if not properties.stanza_ids:
            return

        stanza_id = self._get_stanza_id(properties, archive_jid)
        if stanza_id is None:
            return

        if not self._is_catch_up_finished(archive_jid):
            return

        if timestamp is not None:
            timestamp = datetime.fromtimestamp(timestamp, UTC)

        app.storage.archive.upsert_row(
            mod.MAMArchiveState(
                account_=self._account,
                remote_jid_=JID.from_string(archive_jid),
                to_stanza_id=stanza_id.id,
                to_stanza_ts=timestamp,
            )
        )

    def _mam_message_received(self,
                              _con: types.NBXMPPClient,
                              stanza: Message,
                              properties: MessageProperties
                              ) -> None:

        if not properties.is_mam_message:
            return

        app.ged.raise_event(
            RawMamMessageReceived(account=self._account,
                                  stanza=stanza,
                                  properties=properties))

        assert properties.mam is not None
        if not self._from_valid_archive(stanza, properties):
            self._log.warning('Message from invalid archive %s',
                              properties.mam.archive)
            raise nbxmpp.NodeProcessed

        self._log.info('Received message from archive: %s',
                       properties.mam.archive)
        if not self._is_valid_request(properties):
            self._log.warning('Invalid MAM Message: unknown query id %s',
                              properties.mam.query_id)
            self._log.warning(stanza)
            raise nbxmpp.NodeProcessed

        assert properties.remote_jid is not None
        stanza_id = properties.mam.id

        if app.storage.archive.check_if_stanza_id_exists(
            self._account,
            properties.remote_jid,
            stanza_id
        ):
            self._log.info('Received duplicated message from MAM: %s', stanza_id)
            raise nbxmpp.NodeProcessed

    def _is_valid_request(self, properties: MessageProperties) -> bool:
        assert properties.mam is not None
        valid_id = self._mam_query_ids.get(properties.mam.archive, None)
        return valid_id == properties.mam.query_id

    def _get_query_id(self, jid: JID) -> str:
        query_id = generate_id()
        self._mam_query_ids[jid] = query_id
        return query_id

    def _get_query_params(self) -> tuple[str | None, datetime | None]:
        own_jid = self._con.get_own_jid().new_as_bare()
        archive = app.storage.archive.get_mam_archive_state(
            self._account, own_jid)

        mam_id = None
        if archive is not None:
            mam_id = archive.to_stanza_id

        start_date = None
        if mam_id:
            self._log.info('Request archive: %s, after mam-id %s',
                           own_jid, mam_id)

        else:
            # First Start, we request the last week
            start_date = datetime.now(UTC) - timedelta(days=7)
            self._log.info('Request archive: %s, after date %s',
                           own_jid, start_date)
        return mam_id, start_date

    def _get_muc_query_params(self,
                              jid: JID,
                              threshold: int
                              ) -> tuple[str | None, datetime | None]:

        archive = app.storage.archive.get_mam_archive_state(self._account, jid)
        mam_id = None
        start_date = None
        now = datetime.now(UTC)

        if archive is None or archive.to_stanza_id is None:
            # First join
            start_date = now - timedelta(days=1)
            self._log.info('Request archive: %s, after date %s',
                           jid, start_date)

        elif threshold == SyncThreshold.NO_THRESHOLD:
            # Not our first join and no threshold set

            mam_id = archive.to_stanza_id
            self._log.info('Request archive: %s, after mam-id %s',
                           jid, archive.to_stanza_id)

        else:
            # Not our first join, check how much time elapsed since our
            # last join and check against threshold
            last = archive.to_stanza_ts
            if last is None:
                self._log.info('No last muc timestamp found: %s', jid)
                last = FIRST_UTC_DATETIME

            if now - last > timedelta(days=threshold):
                # To much time has elapsed since last join, apply threshold
                start_date = now - timedelta(days=threshold)
                self._log.info('Too much time elapsed since last join, '
                               'request archive: %s, after date %s, '
                               'threshold: %s', jid, start_date, threshold)

            else:
                # Request from last mam-id
                mam_id = archive.to_stanza_id
                self._log.info('Request archive: %s, after mam-id %s:',
                               jid, archive.to_stanza_id)

        return mam_id, start_date

    @as_task
    def request_archive_on_signin(self) -> Generator[Any, Any]:
        _task = yield  # noqa: F841

        own_jid = self._con.get_own_jid().new_as_bare()

        if own_jid in self._mam_query_ids:
            self._log.warning('request already running for %s', own_jid)
            return

        mam_id, start_date = self._get_query_params()

        result = yield self._execute_query(own_jid, mam_id, start_date)
        if is_error(result):
            if result.condition != 'item-not-found':
                self._log.warning(result)
                return

            self._log.warning(result)
            self._log.warning('Reset archive state: %s', own_jid)
            app.storage.archive.reset_mam_archive_state(
                self._account, result.jid)
            _, start_date = self._get_query_params()
            result = yield self._execute_query(result.jid, None, start_date)
            if is_error(result):
                self._log.warning(result)
                return

        assert isinstance(result, MAMQueryData)
        if result.rsm.last is not None:
            # <last> is not provided if the requested page was empty
            # so this means we did not get anything hence we only need
            # to update the archive info if <last> is present
            app.storage.archive.upsert_row(
                mod.MAMArchiveState(
                    account_=self._account,
                    remote_jid_=result.jid,
                    to_stanza_id=result.rsm.last,
                    to_stanza_ts=datetime.now(UTC),
                )
            )

        if start_date is not None:
            # Record the earliest timestamp we request from
            # the account archive. For the account archive we only
            # set start_date at the very first request.
            app.storage.archive.upsert_row(
                mod.MAMArchiveState(
                    account_=self._account,
                    remote_jid_=result.jid,
                    from_stanza_ts=start_date
                )
            )

    @as_task
    def request_archive_on_muc_join(self, jid: JID) -> Generator[Any, Any]:
        _task = yield  # noqa: F841

        threshold = app.settings.get_group_chat_setting(self._account,
                                                        jid,
                                                        'sync_threshold')
        self._log.info('Threshold for %s: %s', jid, threshold)

        if threshold == SyncThreshold.NO_SYNC:
            return

        contact = self._get_contact_if_exists(jid)
        if contact is not None:
            contact.notify('mam-sync-started')

        mam_id, start_date = self._get_muc_query_params(jid, threshold)

        result = yield self._execute_query(jid, mam_id, start_date)
        if is_error(result):
            if result.condition != 'item-not-found':
                self._log.warning(result)
                if contact is not None:
                    contact.notify('mam-sync-error', result.get_text())
                return

            self._log.warning(result)
            self._log.warning('Reset archive state: %s', jid)
            app.storage.archive.reset_mam_archive_state(
                self._account, result.jid)
            _, start_date = self._get_muc_query_params(jid, threshold)
            result = yield self._execute_query(result.jid, None, start_date)
            if is_error(result):
                self._log.warning(result)
                if contact is not None:
                    contact.notify('mam-sync-error', result.get_text())
                return

        assert isinstance(result, MAMQueryData)
        if result.rsm.last is not None:
            # <last> is not provided if the requested page was empty
            # so this means we did not get anything hence we only need
            # to update the archive info if <last> is present
            app.storage.archive.upsert_row(
                mod.MAMArchiveState(
                    account_=self._account,
                    remote_jid_=result.jid,
                    to_stanza_id=result.rsm.last,
                    to_stanza_ts=datetime.now(UTC)
                )
            )

        if contact is not None:
            contact.notify('mam-sync-finished')

    @as_task
    def _execute_query(self,
                       jid: JID,
                       mam_id: str | None,
                       start_date: datetime | None
                       ) -> Generator[Any, Any]:
        _task = yield  # noqa: F841

        if jid in self._catch_up_finished:
            self._catch_up_finished.remove(jid)

        queryid = self._get_query_id(jid)

        result = yield self.make_query(jid,
                                       queryid,
                                       after=mam_id,
                                       start=start_date)

        self._remove_query_by_jid(result.jid)

        raise_if_error(result)

        assert isinstance(result, MAMQueryData)
        while not result.complete:
            app.storage.archive.upsert_row(
                mod.MAMArchiveState(
                    account_=self._account,
                    remote_jid_=result.jid,
                    to_stanza_id=result.rsm.last,
                )
            )

            queryid = self._get_query_id(result.jid)

            result = yield self.make_query(result.jid,
                                           queryid,
                                           after=result.rsm.last,
                                           start=start_date)

            self._remove_query_by_jid(result.jid)

            raise_if_error(result)

        self._catch_up_finished.append(result.jid)
        self._log.info('Request finished: %s, last mam id: %s',
                       result.jid, result.rsm.last)
        yield result

    def request_archive_interval(self,
                                 start_date: datetime,
                                 end_date: datetime,
                                 after: str | None = None,
                                 queryid: str | None = None
                                 ) -> str:

        jid = self._con.get_own_jid().new_as_bare()

        if after is None:
            self._log.info('Request interval: %s, from %s to %s',
                           jid, start_date, end_date)
        else:
            self._log.info('Request page: %s, after %s', jid, after)

        if queryid is None:
            queryid = self._get_query_id(jid)
        self._mam_query_ids[jid] = queryid

        self.make_query(jid,
                        queryid,
                        after=after,
                        start=start_date,
                        end=end_date,
                        callback=self._on_interval_result,
                        user_data=(queryid, start_date, end_date))
        return queryid

    def _on_interval_result(self, task: Task) -> None:
        queryid, start_date, end_date = task.get_user_data()

        try:
            result = task.finish()
        except (StanzaError, MalformedStanzaError):
            self._remove_query_by_id(queryid)
            return

        self._remove_query_by_jid(result.jid)

        if start_date:
            timestamp = start_date
        else:
            timestamp = ArchiveState.ALL

        assert isinstance(result, MAMQueryData)
        if result.complete:
            self._log.info('Request finished: %s, last mam id: %s',
                           result.jid, result.rsm.last)
            app.storage.archive.upsert_row(
                mod.MAMArchiveState(
                    account_=self._account,
                    remote_jid_=result.jid,
                    from_stanza_ts=timestamp,
                )
            )

            app.ged.raise_event(ArchivingIntervalFinished(
                account=self._account,
                query_id=queryid))

        else:
            self.request_archive_interval(start_date,
                                          end_date,
                                          result.rsm.last,
                                          queryid)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import Gtk

from gajim.common.i18n import _
from gajim.common.styling import ParsingResult
from gajim.common.styling import PlainBlock
from gajim.common.styling import PreBlock
from gajim.common.styling import process
from gajim.common.styling import QuoteBlock

from gajim.gtk.const import MAX_MESSAGE_LENGTH
from gajim.gtk.conversation.code_widget import CodeWidget
from gajim.gtk.conversation.plain_widget import PlainWidget
from gajim.gtk.conversation.quote_widget import QuoteWidget
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import container_remove_all
from gajim.gtk.util.misc import iterate_children

ContentT = ParsingResult | QuoteBlock


class MessageWidget(Gtk.Box, SignalManager):
    def __init__(self, account: str, selectable: bool = True) -> None:
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL)
        SignalManager.__init__(self)

        self._account = account
        self._selectable = selectable

        self._content: ContentT | None = None
        self._nickname = None
        self._original_text = ""
        self._action_phrase_text = ""

    def do_unroot(self) -> None:
        self._disconnect_all()
        Gtk.Box.do_unroot(self)

    def get_content(self) -> ContentT | None:
        return self._content

    def get_text(self) -> str:
        if self._content is not None:
            return self._original_text
        return self._action_phrase_text

    def set_selectable(self, selectable: bool) -> None:
        for widget in iterate_children(self):
            if isinstance(widget, PlainWidget):
                widget.set_selectable(selectable)

    def add_with_styling(
        self, text: str, nickname: str | None = None, show_full_text: bool = False
    ) -> None:
        self._original_text = text
        self._nickname = nickname

        text_over_max = len(text) > MAX_MESSAGE_LENGTH and not show_full_text

        if text_over_max:
            text = text[:MAX_MESSAGE_LENGTH]

        if text.startswith("/me ") and nickname is not None:
            self._add_action_phrase(text, nickname)
        else:
            result = process(text)
            self.add_content(result)

        if text_over_max:
            self._add_read_more_button()

    def _add_action_phrase(self, text: str, nickname: str):
        self.clear()
        widget = PlainWidget(self._account, self._selectable)
        widget.add_action_phrase(text, nickname)
        self.append(widget)

        self._action_phrase_text = text.replace("/me", f"* {nickname}", 1)

    def add_content(self, content: ContentT) -> None:
        self.clear()
        self._content = content
        for block in content.blocks:
            if isinstance(block, PlainBlock):
                widget = PlainWidget(self._account, self._selectable)
                widget.add_content(block)
                self.append(widget)
                continue

            if isinstance(block, PreBlock):
                widget = CodeWidget(self._account)
                widget.add_content(block)
                self.append(widget)
                continue

            if isinstance(block, QuoteBlock):
                message_widget = MessageWidget(self._account, self._selectable)
                message_widget.add_content(block)
                widget = QuoteWidget(self._account)
                widget.append(message_widget)
                self.append(widget)
                continue

    def _add_read_more_button(self) -> None:
        link_button = Gtk.LinkButton(label=_("[read more]"), halign=Gtk.Align.START)
        self._connect(link_button, "activate-link", self._on_read_more)
        self.append(link_button)

    def _on_read_more(self, _button: Gtk.LinkButton) -> bool:
        self.set_can_focus(False)
        self.add_with_styling(self._original_text, self._nickname, True)
        GLib.idle_add(self.set_can_focus, True)
        return Gdk.EVENT_STOP

    def clear(self) -> None:
        container_remove_all(self)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from datetime import datetime

from gi.repository import Gtk

from gajim.common.const import AvatarSize
from gajim.common.util.datetime import utc_now
from gajim.common.util.text import process_non_spacing_marks

from gajim.gtk.conversation.rows.base import BaseRow
from gajim.gtk.conversation.rows.widgets import DateTimeLabel
from gajim.gtk.conversation.rows.widgets import SimpleLabel


class InfoMessage(BaseRow):
    def __init__(self, account: str, text: str, timestamp: datetime | None) -> None:
        BaseRow.__init__(self, account)

        self.type = "info"

        if timestamp is None:
            timestamp = utc_now()
        self.timestamp = timestamp.astimezone()
        self.db_timestamp = timestamp.timestamp()

        avatar_placeholder = Gtk.Box()
        avatar_placeholder.set_size_request(AvatarSize.ROSTER, -1)
        self.grid.attach(avatar_placeholder, 0, 0, 1, 1)

        icon = Gtk.Image.new_from_icon_name("lucide-info-symbolic")
        icon.add_css_class("gajim-status-message")
        self.grid.attach(icon, 1, 0, 1, 1)

        self._label = SimpleLabel()
        self._label.add_css_class("gajim-status-message")
        self._label.set_text(process_non_spacing_marks(text))
        self.grid.attach(self._label, 2, 0, 1, 1)

        timestamp_widget = DateTimeLabel(self.timestamp)
        timestamp_widget.set_halign(Gtk.Align.START)
        timestamp_widget.set_valign(Gtk.Align.END)
        self.grid.attach(timestamp_widget, 3, 0, 1, 1)

    def do_unroot(self) -> None:
        BaseRow.do_unroot(self)
